"""Window module handles window settings.

Notes:
    Thank you to `honeybee-energy <https://github.com/ladybug-tools/honeybee-energy/
    blob/master/honeybee_energy/construction/window.py>`_ for implementing center
    of glass resistance formulas from ISO. Those where adapted to the structure of the
    archetypal.template module.
"""

import collections
from enum import Enum

from validator_collection import validators

from archetypal.simple_glazing import calc_simple_glazing
from archetypal.template.constructions.base_construction import LayeredConstruction
from archetypal.template.materials.gas_layer import GasLayer
from archetypal.template.materials.gas_material import GasMaterial
from archetypal.template.materials.glazing_material import GlazingMaterial
from archetypal.template.materials.material_layer import MaterialLayer


class WindowType(Enum):
    """Refers to the window type. Two choices are available: interior or exterior."""

    External = 0
    Internal = 1

    def __lt__(self, other):
        """Return true if self lower than other."""
        return self._value_ < other._value_

    def __gt__(self, other):
        """Return true if self higher than other."""
        return self._value_ > other._value_


class ShadingType(Enum):
    """Refers to window shading types.

    Hint:
        EnergyPlus specifies 8 different shading types, but only 2 are supported
        here: InteriorShade and ExteriorShade. See shading_ for more info.

    .. _shading: https://bigladdersoftware.com/epx/docs/8-4/input-output-reference/group-thermal-zone-description-geometry.html#field-shading-type
    """

    ExteriorShade = 0
    InteriorShade = 1

    def __lt__(self, other):
        """Return true if self lower than other."""
        return self._value_ < other._value_

    def __gt__(self, other):
        """Return true if self higher than other."""
        return self._value_ > other._value_


class WindowConstruction(LayeredConstruction):
    """Window Construction.

    .. image:: ../images/template/constructions-window.png
    """

    _CATEGORIES = ("single", "double", "triple", "quadruple")

    __slots__ = ("_category",)

    def __init__(self, Name, Layers, Category="Double", **kwargs):
        """Initialize a WindowConstruction.

        Args:
            Name (str): Name of the WindowConstruction.
            Layers (list of (MaterialLayer or GasLayer)): List of MaterialLayer and
                GasLayer.
            Category (str): "Single", "Double" or "Triple".
            **kwargs: Other keywords passed to the constructor.
        """
        super(WindowConstruction, self).__init__(
            Name,
            Layers,
            Category=Category,
            **kwargs,
        )
        self.Category = Category  # set here for validators

    @property
    def Category(self):
        """Get or set the Category. Choices are ("single", "double", "triple")."""
        return self._category

    @Category.setter
    def Category(self, value):
        assert value.lower() in self._CATEGORIES, (
            f"Input error for value '{value}'. The "
            f"Category must be one of ({self._CATEGORIES})"
        )
        self._category = value

    @property
    def gap_count(self):
        """Get the number of gas gaps contained within the window construction."""
        count = 0
        for layer in self.Layers:
            if isinstance(layer, GasLayer):
                count += 1
        return count

    @property
    def glazing_count(self):
        """Get the nb of glazing materials contained within the window construction."""
        count = 0
        for layer in self.Layers:
            if isinstance(layer, MaterialLayer):
                count += 1
        return count

    @property
    def r_factor(self):
        """Get the construction R-factor [m2-K/W].

        Note: including standard resistances for air films. Formulas for film
        coefficients come from EN673 / ISO10292.
        """
        gap_count = self.gap_count
        if gap_count == 0:  # single pane
            return (
                self.Layers[0].r_value
                + (1 / self.out_h_simple())
                + (1 / self.in_h_simple())
            )
        elif gap_count == 1:
            heat_transfers, temperature_profile = self.heat_balance("summer", 0)
            *_, Q_dot_i4 = heat_transfers
            return (temperature_profile[-1] - temperature_profile[0]) / Q_dot_i4
        r_vals, emissivities = self._layered_r_value_initial(gap_count)
        r_vals = self._solve_r_values(r_vals, emissivities)
        return sum(r_vals)

    @property
    def r_value(self):
        """Get or set the thermal resistance [K⋅m2/W] (excluding air films)."""
        gap_count = self.gap_count
        if gap_count == 0:  # single pane
            return self.Layers[0].r_value
        r_vals, emissivities = self._layered_r_value_initial(gap_count)
        r_vals = self._solve_r_values(r_vals, emissivities)
        return sum(r_vals[1:-1])

    @property
    def outside_emissivity(self):
        """Get the hemispherical emissivity of the outside face of the construction."""
        return self.Layers[0].Material.IREmissivityFront

    @property
    def inside_emissivity(self):
        """Get the hemispherical emissivity of the inside face of the construction."""
        return self.Layers[-1].Material.IREmissivityBack

    @property
    def solar_transmittance(self):
        """Get the solar transmittance of the window at normal incidence."""
        if self.glazing_count == 2:
            tau_1 = self.Layers[0].Material.SolarTransmittance
            tau_2 = self.Layers[-1].Material.SolarTransmittance
            rho_1 = self.Layers[0].Material.SolarReflectanceFront
            rho_2 = self.Layers[-1].Material.SolarReflectanceFront
            return (tau_1 * tau_2) / (1 - rho_1 * rho_2)
        trans = 1
        for layer in self.Layers:
            if isinstance(layer.Material, GlazingMaterial):
                trans *= layer.Material.SolarTransmittance
        return trans

    @property
    def visible_transmittance(self):
        """Get the visible transmittance of the window at normal incidence."""
        trans = 1
        for layer in self.Layers:
            if isinstance(layer.Material, GlazingMaterial):
                trans *= layer.Material.VisibleTransmittance
        return trans

    @property
    def thickness(self):
        """Thickness of the construction [m]."""
        thickness = 0
        for layer in self.Layers:
            thickness += layer.Thickness
        return thickness

    @classmethod
    def from_dict(cls, data, materials, **kwargs):
        """Create an WindowConstruction from a dictionary.

        Args:
            data (dict): The python dictionary.
            materials (dict): A dictionary of materials with their id as keys.
            **kwargs: keywords passed to the constructor.

        .. code-block:: python

            data = {
                "$id": "57",
                "Layers": [
                    {"Material": {"$ref": "7"}, "Thickness": 0.003},
                    {"Material": {"$ref": "1"}, "Thickness": 0.006},
                    {"Material": {"$ref": "7"}, "Thickness": 0.003},
                ],
                "AssemblyCarbon": 0.0,
                "AssemblyCost": 0.0,
                "AssemblyEnergy": 0.0,
                "DisassemblyCarbon": 0.0,
                "DisassemblyEnergy": 0.0,
                "Category": "Double",
                "Comments": "default",
                "DataSource": "default",
                "Name": "B_Dbl_Air_Cl",
            }
        """
        _id = data.pop("$id")
        layers = [
            MaterialLayer(materials[layer["Material"]["$ref"]], layer["Thickness"])
            if isinstance(
                materials[layer["Material"]["$ref"]], (MaterialLayer, GlazingMaterial)
            )
            else GasLayer(materials[layer["Material"]["$ref"]], layer["Thickness"])
            for layer in data.pop("Layers")
        ]
        return cls(Layers=layers, id=_id, **data, **kwargs)

    @classmethod
    def from_epbunch(cls, Construction, **kwargs):
        """Create :class:`WindowConstruction` object from idf Construction object.

        Example:
            >>> from archetypal import IDF
            >>> from archetypal.template.window_setting import WindowSetting
            >>> idf = IDF("myidf.idf")
            >>> construction_name = "Some construction name"
            >>> WindowConstruction.from_epbunch(Name=construction_name, idf=idf)

        Args:
            Construction (EpBunch): The Construction epbunch object.
            **kwargs: Other keywords passed to the constructor.
        """
        layers = WindowConstruction._layers_from_construction(Construction, **kwargs)
        catdict = {0: "Single", 1: "Single", 2: "Double", 3: "Triple", 4: "Quadruple"}
        category = catdict[
            len([lyr for lyr in layers if isinstance(lyr.Material, GlazingMaterial)])
        ]
        return cls(Name=Construction.Name, Layers=layers, Category=category, **kwargs)

    @classmethod
    def from_shgc(
        cls,
        Name,
        solar_heat_gain_coefficient,
        u_factor,
        visible_transmittance=None,
        **kwargs,
    ):
        """Create a WindowConstruction from shgc, u_factor and visible_transmittance.

        Args:
            Name (str): The name of the window construction.
            shgc (double): The window's Solar Heat Gain Coefficient.
            u_factor (double): The window's U-value.
            visible_transmittance (double, optional): The window's visible
                transmittance. If none, the visible transmittance defaults to the
                solar transmittance t_sol.
            kwargs: keywrods passed to the parent constructor.

        Returns:

        """
        glass_properties = calc_simple_glazing(
            solar_heat_gain_coefficient,
            u_factor,
            visible_transmittance,
        )
        material_obj = GlazingMaterial(Name="Simple Glazing", **glass_properties)

        material_layer = MaterialLayer(material_obj, glass_properties["Thickness"])
        return cls(Name, Layers=[material_layer], **kwargs)

    def to_dict(self):
        """Return WindowConstruction dictionary representation."""
        self.validate()  # Validate object before trying to get json format

        data_dict = collections.OrderedDict()

        data_dict["$id"] = str(self.id)
        data_dict["Layers"] = [layer.to_dict() for layer in self.Layers]
        data_dict["AssemblyCarbon"] = self.AssemblyCarbon
        data_dict["AssemblyCost"] = self.AssemblyCost
        data_dict["AssemblyEnergy"] = self.AssemblyEnergy
        data_dict["DisassemblyCarbon"] = self.DisassemblyCarbon
        data_dict["DisassemblyEnergy"] = self.DisassemblyEnergy
        data_dict["Category"] = self.Category
        data_dict["Comments"] = validators.string(self.Comments, allow_empty=True)
        data_dict["DataSource"] = self.DataSource
        data_dict["Name"] = self.Name

        return data_dict

    def to_epbunch(self, idf):
        """Convert self to a `Construction` epbunch given an idf model.

        Args:
            idf (IDF): The idf model in which the EpBunch is created.

        .. code-block:: python

            Construction,
                B_Dbl_Air_Cl,                           !- Name
                B_Glass_Clear_3_0.003_B_Dbl_Air_Cl,     !- Outside Layer
                AIR_0.006_B_Dbl_Air_Cl,                 !- Layer 2
                B_Glass_Clear_3_0.003_B_Dbl_Air_Cl;     !- Layer 3

        Returns:
            EpBunch: The EpBunch object added to the idf model.
        """
        data = {"Name": self.Name}
        for i, layer in enumerate(self.Layers):
            mat = layer.to_epbunch(idf)
            if i < 1:
                data["Outside_Layer"] = mat.Name
            else:
                data[f"Layer_{i+1}"] = mat.Name

        return idf.newidfobject("CONSTRUCTION", **data)

    def mapping(self, validate=True):
        """Get a dict based on the object properties, useful for dict repr.

        Args:
            validate (bool): If True, try to validate object before returning the
                mapping.
        """
        if validate:
            self.validate()

        return dict(
            Layers=self.Layers,
            AssemblyCarbon=self.AssemblyCarbon,
            AssemblyCost=self.AssemblyCost,
            AssemblyEnergy=self.AssemblyEnergy,
            DisassemblyCarbon=self.DisassemblyCarbon,
            DisassemblyEnergy=self.DisassemblyEnergy,
            Category=self.Category,
            Comments=self.Comments,
            DataSource=self.DataSource,
            Name=self.Name,
        )

    def combine(self, other, weights=None):
        """Append other to self. Return self + other as a new object.

        For now, simply returns self.

        todo:
            - Implement equivalent window layers for constant u-factor.

        """
        # Check if other is None. Simply return self
        if not other:
            return self

        if not self:
            return other

        return self

    def validate(self):
        """Validate object and fill in missing values.

        todo:
            - Implement validation
        """
        return self

    def duplicate(self):
        """Get copy of self."""
        return self.__copy__()

    def temperature_profile(
        self,
        outside_temperature=-18,
        inside_temperature=21,
        wind_speed=6.7,
        height=1.0,
        angle=90.0,
        pressure=101325,
    ):
        """Get a list of temperatures at each material boundary across the construction.

        Args:
            outside_temperature: The temperature on the outside of the construction [C].
                Default is -18, which is consistent with NFRC 100-2010.
            inside_temperature: The temperature on the inside of the construction [C].
                Default is 21, which is consistent with NFRC 100-2010.
            wind_speed: The average outdoor wind speed [m/s]. This affects outdoor
                convective heat transfer coefficient. Default is 6.7 m/s.
            height: An optional height for the surface in meters. Default is 1.0 m.
            angle: An angle in degrees between 0 and 180.
                0 = A horizontal surface with the outside boundary on the bottom.
                90 = A vertical surface
                180 = A horizontal surface with the outside boundary on the top.
            pressure: The average pressure of in Pa.
                Default is 101325 Pa for standard pressure at sea level.
        Returns:
            A tuple with two elements
            -   temperatures: A list of temperature values [C].
                The first value will always be the outside temperature and the
                second will be the exterior surface temperature.
                The last value will always be the inside temperature and the second
                to last will be the interior surface temperature.
            -   r_values: A list of R-values for each of the material layers [m2-K/W].
                The first value will always be the resistance of the exterior air
                and the last value is the resistance of the interior air.
                The sum of this list is the R-factor for this construction given
                the input parameters.
        """
        # reverse the angle if the outside temperature is greater than the inside one
        if angle != 90 and outside_temperature > inside_temperature:
            angle = abs(180 - angle)
        gap_count = self.gap_count

        # single pane or simple glazing system
        if gap_count == 0:
            in_r_init = 1 / self.in_h_simple()
            r_values = [
                1 / self.out_h(wind_speed, outside_temperature + 273.15),
                self.Layers[0].r_value,
                in_r_init,
            ]
            in_delta_t = (in_r_init / sum(r_values)) * (
                outside_temperature - inside_temperature
            )
            r_values[-1] = 1 / self.in_h(
                inside_temperature - (in_delta_t / 2) + 273.15,
                in_delta_t,
                height,
                angle,
                pressure,
            )
            temperatures = self._temperature_profile_from_r_values(
                r_values, outside_temperature, inside_temperature
            )
            return temperatures, r_values

        # multi-layered window construction
        guess = abs(inside_temperature - outside_temperature) / 2
        guess = 1 if guess < 1 else guess  # prevents zero division with gas conductance
        avg_guess = ((inside_temperature + outside_temperature) / 2) + 273.15
        r_values, emissivities = self._layered_r_value_initial(
            gap_count, guess, avg_guess, wind_speed
        )
        r_last = 0
        r_next = sum(r_values)
        while abs(r_next - r_last) > 0.001:  # 0.001 is the r-value tolerance
            r_last = sum(r_values)
            temperatures = self._temperature_profile_from_r_values(
                r_values, outside_temperature, inside_temperature
            )
            r_values = self._layered_r_value(
                temperatures, r_values, emissivities, height, angle, pressure
            )
            r_next = sum(r_values)
        temperatures = self._temperature_profile_from_r_values(
            r_values, outside_temperature, inside_temperature
        )
        return temperatures, r_values

    @staticmethod
    def _layers_from_construction(construction, **kwargs):
        """Retrieve layers for the Construction epbunch."""
        layers = []
        for field in construction.fieldnames[2:]:
            # Loop through the layers from the outside layer towards the
            # indoor layers and get the material they are made of.
            material = construction.get_referenced_object(field) or kwargs.get(
                "material", None
            )
            if material:
                # Create the WindowMaterial:Glazing or the WindowMaterial:Gas
                # and append to the list of layers
                if material.key.upper() == "WindowMaterial:Glazing".upper():
                    material_obj = GlazingMaterial(
                        Name=material.Name,
                        Conductivity=material.Conductivity,
                        SolarTransmittance=material.Solar_Transmittance_at_Normal_Incidence,
                        SolarReflectanceFront=material.Front_Side_Solar_Reflectance_at_Normal_Incidence,
                        SolarReflectanceBack=material.Back_Side_Solar_Reflectance_at_Normal_Incidence,
                        VisibleTransmittance=material.Visible_Transmittance_at_Normal_Incidence,
                        VisibleReflectanceFront=material.Front_Side_Visible_Reflectance_at_Normal_Incidence,
                        VisibleReflectanceBack=material.Back_Side_Visible_Reflectance_at_Normal_Incidence,
                        IRTransmittance=material.Infrared_Transmittance_at_Normal_Incidence,
                        IREmissivityFront=material.Front_Side_Infrared_Hemispherical_Emissivity,
                        IREmissivityBack=material.Back_Side_Infrared_Hemispherical_Emissivity,
                        DirtFactor=material.Dirt_Correction_Factor_for_Solar_and_Visible_Transmittance,
                        Optical=material.Optical_Data_Type,
                        OpticalData=material.Window_Glass_Spectral_Data_Set_Name,
                    )

                    material_layer = MaterialLayer(material_obj, material.Thickness)

                elif material.key.upper() == "WindowMaterial:Gas".upper():
                    # Todo: Make gas name generic, like in UmiTemplateLibrary Editor
                    material_obj = GasMaterial(
                        Name=material.Gas_Type.upper(), Conductivity=0.02
                    )
                    material_layer = GasLayer(material_obj, material.Thickness)
                elif material.key.upper() == "WINDOWMATERIAL:SIMPLEGLAZINGSYSTEM":
                    glass_properties = calc_simple_glazing(
                        material.Solar_Heat_Gain_Coefficient,
                        material.UFactor,
                        material.Visible_Transmittance,
                    )
                    material_obj = GlazingMaterial(
                        Name=material.Name, **glass_properties
                    )

                    material_layer = MaterialLayer(
                        material_obj, glass_properties["Thickness"]
                    )
                    layers.append(material_layer)
                    break
                else:
                    continue

                layers.append(material_layer)
        return layers

    def _layered_r_value_initial(
        self, gap_count, delta_t_guess=15, avg_t_guess=273.15, wind_speed=6.7
    ):
        """Compute initial r-values of each layer within a layered construction."""
        r_vals = [1 / self.out_h(wind_speed, avg_t_guess - delta_t_guess)]
        emiss = []
        delta_t = delta_t_guess / gap_count
        for i, lyr in enumerate(self.Layers):
            if isinstance(lyr, MaterialLayer):
                r_vals.append(lyr.r_value)
                emiss.append(None)
            else:  # gas layer
                e_front = self.Layers[i + 1].Material.IREmissivityFront
                e_back = self.Layers[i + 1].Material.IREmissivityBack
                r_vals.append(
                    1 / lyr.u_value(delta_t, e_back, e_front, t_kelvin=avg_t_guess)
                )
                emiss.append((e_back, e_front))
        r_vals.append(1 / self.in_h_simple())
        return r_vals, emiss

    def __hash__(self):
        """Return the hash value of self."""
        return hash((self.__class__.__name__, getattr(self, "Name", None)))

    def __eq__(self, other):
        """Assert self is equivalent to other."""
        if not isinstance(other, WindowConstruction):
            return NotImplemented
        else:
            return all(
                [
                    self.Category == other.Category,
                    self.AssemblyCarbon == other.AssemblyCarbon,
                    self.AssemblyCost == other.AssemblyCost,
                    self.AssemblyEnergy == other.AssemblyEnergy,
                    self.DisassemblyCarbon == other.DisassemblyCarbon,
                    self.DisassemblyEnergy == other.DisassemblyEnergy,
                    self.Layers == other.Layers,
                ]
            )

    def __copy__(self):
        """Create a copy of self."""
        return self.__class__(**self.mapping())

    def __add__(self, other):
        """Combine self and other."""
        return self.combine(other)

    def _solve_r_values(self, r_values, emissivities):
        """Solve iteratively for R-values."""
        r_last = 0
        r_next = sum(r_values)
        while abs(r_next - r_last) > 0.001:  # 0.001 is the r-value tolerance
            r_last = sum(r_values)
            temperatures = self._temperature_profile_from_r_values(r_values)
            r_values = self._layered_r_value(temperatures, r_values, emissivities)
            r_next = sum(r_values)
        return r_values

    def _temperature_profile_from_r_values(
        self, r_values, outside_temperature=-18, inside_temperature=21
    ):
        """Get a list of temperatures at each material boundary between R-values."""
        r_factor = sum(r_values)
        delta_t = inside_temperature - outside_temperature
        temperatures = [outside_temperature]
        for i, r_val in enumerate(r_values):
            temperatures.append(temperatures[i] + (delta_t * (r_val / r_factor)))
        return temperatures

    def _layered_r_value(
        self,
        temperatures,
        r_values_init,
        emiss,
        height=1.0,
        angle=90.0,
        pressure=101325,
    ):
        """Compute delta_t adjusted r-values of each layer within a construction."""
        r_vals = [r_values_init[0]]
        for i, layer in enumerate(self.Layers):
            if isinstance(layer, MaterialLayer):
                r_vals.append(r_values_init[i + 1])
            elif isinstance(layer, GasLayer):  # gas layer
                delta_t = abs(temperatures[i + 1] - temperatures[i + 2])
                avg_temp = ((temperatures[i + 1] + temperatures[i + 2]) / 2) + 273.15
                r_vals.append(
                    1
                    / layer.u_value_at_angle(
                        delta_t,
                        emiss[i][0],
                        emiss[i][1],
                        height,
                        angle,
                        avg_temp,
                        pressure,
                    )
                )
        delta_t = abs(temperatures[-1] - temperatures[-2])
        avg_temp = ((temperatures[-1] + temperatures[-2]) / 2) + 273.15
        r_vals.append(1 / self.in_h(avg_temp, delta_t, height, angle, pressure))
        return r_vals

    def shgc(self, environmental_conditions="summer", global_radiation=783):
        """Calculate the shgc given environmental conditions.

        Notes:
            This method implements a heat balance at each interface of the
            glazing unit including outside and inside air film resistances,
            solar radiation absorption in the glass. See
            :meth:`~archetypal.template.constructions.window_construction
            .WindowConstruction.heat_balance` for more details.

        Args:
            environmental_conditions (str): "summer" or "winter". A window shgc is
                usually calculated with summer conditions. Default is "summer".
            global_radiation (float): Incident solar radiation [W / m ^ 2]. Overwrite
                the solar radiation used in the calculation of the shgc.

        Returns:
            float: The shgc of the window construction for the given environmental
            conditions.
        """
        # Q_dot_noSun
        heat_transfers, temperature_profile = self.heat_balance(
            environmental_conditions, 0
        )
        *_, Q_dot_noSun = heat_transfers

        # Q_dot_Sun
        heat_transfers, temperature_profile = self.heat_balance(
            environmental_conditions, 783
        )
        *_, Q_dot_i4 = heat_transfers

        Q_dot_sun = -Q_dot_i4 + self.solar_transmittance * global_radiation
        shgc = (Q_dot_sun - -Q_dot_noSun) / global_radiation
        return shgc

    def heat_balance(self, environmental_conditions="summer", G_t=783):
        """Return heat flux and temperatures at each surface of the window.

        Note: Only implemented for glazing with two layers.

        Args:
            environmental_conditions (str): The environmental conditions from
                the NRFC standard used to calculate the heat balance. Default is
                "summer".
            G_t (float): The incident radiation.

        Returns:
            tuple: heat_flux, temperature_profile
        """
        assert (
            self.glazing_count == 2
        ), f"Expected a window with 2 glazing layers, not {self.glazing_count}."
        ENV_CONDITIONS = {"summer": [32, 24, 2.75], "winter": [-18, 21, 5.5]}
        (
            outside_temperature,
            inside_temperature,
            environmental_conditions,
        ) = ENV_CONDITIONS[environmental_conditions]
        temperatures_next, r_values = self.temperature_profile(
            outside_temperature, inside_temperature, environmental_conditions
        )
        temperatures_last = [0] * 6
        pressure = 101325  # [Pa]
        height = 1  # [m]
        angle = 90  # degree
        k_g = 1  # [W / m - K]
        sigma = 5.670e-8  # [W / m2 - K4]
        # "21011 (SGG Planiclear 4 mm), Air 12 mm, 21414 (SGG Planitherm One 4 mm)"
        epsilon_1 = self.Layers[0].Material.IREmissivityFront  # [-]
        epsilon_2 = self.Layers[0].Material.IREmissivityBack  # [-]
        epsilon_3 = self.Layers[-1].Material.IREmissivityFront  # [-]
        epsilon_4 = self.Layers[-1].Material.IREmissivityBack  # [-]
        abs_1 = 0.0260  # [-]
        abs_2 = 0.0737  # [-]
        L_12 = self.Layers[0].Thickness  # [m]
        L_23 = self.Layers[1].Thickness  # [m]  # included in Layers[1] used below
        L_34 = self.Layers[2].Thickness  # [m]
        # "Solar absorption distribution"
        Q_dot_abs_1 = abs_1 / self.glazing_count * G_t
        Q_dot_abs_2 = abs_1 / self.glazing_count * G_t
        Q_dot_abs_3 = abs_2 / self.glazing_count * G_t
        Q_dot_abs_4 = abs_2 / self.glazing_count * G_t
        T_o, T_1, T_2, T_3, T_4, T_i = temperatures_next
        while not self.assert_almost_equal(temperatures_last, temperatures_next):
            temperatures_last = [T_o, T_1, T_2, T_3, T_4, T_i]

            # "Heat balance at surface 1"
            h_c_o = 4 + 4 * environmental_conditions
            Q_dot_c_1o = h_c_o * (T_1 - T_o)
            Q_dot_r_1o = epsilon_1 * sigma * ((T_1 + 273.15) ** 4 - (T_o + 273.15) ** 4)
            Q_dot_1o = Q_dot_c_1o + Q_dot_r_1o
            Q_dot_21 = k_g / L_12 * (T_2 - T_1)  # + Q_dot_abs_1

            # "Heat balance at surface 2"
            h_c_32 = self.Layers[1].convective_conductance_at_angle(
                abs(T_3 - T_2), height, angle, (T_3 + T_2) / 2 + 273.15, pressure
            )
            Q_dot_c_32 = h_c_32 * (T_3 - T_2)
            Q_dot_r_32 = (
                sigma
                * ((T_3 + 273.15) ** 4 - (T_2 + 273.15) ** 4)
                / (1 / epsilon_2 + 1 / epsilon_3 - 1)
            )
            Q_dot_32 = Q_dot_c_32 + Q_dot_r_32

            # "Heat balance at surface 3"
            Q_dot_43 = k_g / L_34 * (T_4 - T_3)  # + Q_dot_abs_3

            # Q_dot_32 = Q_dot_abs_3 + Q_dot_43

            # "Heat balance at surface 4"
            h_c_i = self.in_h_c(
                (T_4 + T_i) / 2 + 273.15, abs(T_i - T_4), height, angle, pressure
            )
            Q_dot_c_i4 = h_c_i * (T_i - T_4)
            Q_dot_r_i4 = epsilon_4 * sigma * ((T_i + 273.15) ** 4 - (T_4 + 273.15) ** 4)
            Q_dot_i4 = Q_dot_c_i4 + Q_dot_r_i4

            # calc new temps
            T_1 = T_2 - (Q_dot_1o - Q_dot_abs_1) / k_g * L_12
            T_2 = (Q_dot_abs_2 + Q_dot_32) / k_g * L_12 + T_1
            T_3 = T_4 - (Q_dot_32 - Q_dot_abs_3) / k_g * L_34
            T_4 = (Q_dot_abs_4 + Q_dot_i4) / k_g * L_34 + T_3

            temperatures_next = [T_o, T_1, T_2, T_3, T_4, T_i]
        heat_transfers = [Q_dot_1o, Q_dot_21, Q_dot_32, Q_dot_43, Q_dot_i4]
        return heat_transfers, temperatures_next

    def assert_almost_equal(self, temperatures_last, temperatures_next):
        return all(
            abs(desired - actual) < 1.5 * 10 ** (-3)
            for desired, actual in zip(temperatures_last, temperatures_next)
        )



# To use a consistent encoding
import codecs
import os
import re
import sys
from os import path

# Always prefer setuptools over distutils
from setuptools import find_namespace_packages, setup

here = os.getcwd()

# This check is here if the user does not have a new enough pip to recognize
# the minimum Python requirement in the metadata.
if sys.version_info < (3, 7):
    error = """
archetypal 1.1+ does not support Python 2.x, 3.0, 3.1, 3.2, or 3.3.
Python 3.7 and above is required. This may be due to an out of date pip.
Make sure you have pip >= 9.0.1.
"""
    sys.exit(error)


def read(*parts):
    with codecs.open(path.join(here, *parts), "r") as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


# Get the long description from the README file
with codecs.open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

with open(path.join(here, "requirements.txt")) as f:
    requirements_lines = f.readlines()
install_requires = [r.strip() for r in requirements_lines]

with open(path.join(here, "requirements-dev.txt")) as f:
    requirements_lines = f.readlines()
dev_requires = [r.strip() for r in requirements_lines]

setup(
    name="archetypal",
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    packages=find_namespace_packages(
        include=["archetypal", "archetypal.*"], exclude=["tests"]
    ),
    include_package_data=True,
    url="https://github.com/samuelduchesne/archetypal",
    license="MIT License",
    author="Samuel Letellier-Duchesne",
    author_email="samuel.letellier-duchesne@polymtl.ca",
    description="Retrieve, construct, simulate, convert and analyse building archetypes",
    long_description=long_description,
    keywords="Building archetypes",
    python_requires=">=3.7",
    install_requires=install_requires,
    extras_require={"dev": dev_requires},
    test_suite="tests",
    entry_points="""
        [console_scripts]
        archetypal=archetypal.cli:cli
    """,
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 4 - Beta",
        # Indicate who your project is intended for
        "Intended Audience :: Science/Research",
        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: MIT License",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)



"""archetypal DomesticHotWaterSetting."""

import collections
from statistics import mean

import numpy as np
from eppy import modeleditor
from sigfig import round
from validator_collection import validators

from archetypal import settings
from archetypal.template.schedule import UmiSchedule
from archetypal.template.umi_base import UmiBase
from archetypal.utils import log, reduce, timeit


class DomesticHotWaterSetting(UmiBase):
    """Domestic Hot Water settings.

    .. image:: ../images/template/zoneinfo-dhw.png
    """

    __slots__ = (
        "_flow_rate_per_floor_area",
        "_is_on",
        "_water_schedule",
        "_water_supply_temperature",
        "_water_temperature_inlet",
        "_area",
    )

    def __init__(
        self,
        Name,
        WaterSchedule=None,
        IsOn=True,
        FlowRatePerFloorArea=0.03,
        WaterSupplyTemperature=65,
        WaterTemperatureInlet=10,
        area=1,
        **kwargs,
    ):
        """Initialize object with parameters.

        Args:
            area (float): The area the zone associated to this object.
            IsOn (bool): If True, dhw is on.
            WaterSchedule (UmiSchedule): Schedule that modulates the
                FlowRatePerFloorArea.
            FlowRatePerFloorArea (float): The flow rate per flow area [m³/(hr·m²)].
            WaterSupplyTemperature (float): The water supply temperature [degC].
            WaterTemperatureInlet (float): The water temperature intel from the water
                mains [degC].
            **kwargs: keywords passed to parent constructors.
        """
        super(DomesticHotWaterSetting, self).__init__(Name, **kwargs)
        self.FlowRatePerFloorArea = FlowRatePerFloorArea
        self.IsOn = IsOn
        self.WaterSupplyTemperature = WaterSupplyTemperature
        self.WaterTemperatureInlet = WaterTemperatureInlet
        self.WaterSchedule = WaterSchedule
        self.area = area

    @property
    def FlowRatePerFloorArea(self):
        """Get or set the flow rate per flow area [m³/(hr·m²)]."""
        return self._flow_rate_per_floor_area

    @FlowRatePerFloorArea.setter
    def FlowRatePerFloorArea(self, value):
        self._flow_rate_per_floor_area = validators.float(value, minimum=0)

    @property
    def IsOn(self):
        """Get or set the availability of the domestic hot water [bool]."""
        return self._is_on

    @IsOn.setter
    def IsOn(self, value):
        assert isinstance(value, bool), (
            f"Input error with value {value}. IsOn must "
            f"be a boolean, not a {type(value)}"
        )
        self._is_on = value

    @property
    def WaterSchedule(self):
        """Get or set the schedule which modulates the FlowRatePerFloorArea."""
        return self._water_schedule

    @WaterSchedule.setter
    def WaterSchedule(self, value):
        if value is not None:
            assert isinstance(value, UmiSchedule), (
                f"Input error with value {value}. WaterSchedule must "
                f"be an UmiSchedule, not a {type(value)}"
            )
        self._water_schedule = value

    @property
    def WaterSupplyTemperature(self):
        """Get or set the water supply temperature [degC]."""
        return self._water_supply_temperature

    @WaterSupplyTemperature.setter
    def WaterSupplyTemperature(self, value):
        self._water_supply_temperature = validators.float(value)

    @property
    def WaterTemperatureInlet(self):
        """Get or set the water temperature intel from the water mains [degC]."""
        return self._water_temperature_inlet

    @WaterTemperatureInlet.setter
    def WaterTemperatureInlet(self, value):
        self._water_temperature_inlet = validators.float(value)

    @property
    def area(self):
        """Get or set the area of the zone associated to this object [m²]."""
        return self._area

    @area.setter
    def area(self, value):
        self._area = validators.float(value, minimum=0)

    @classmethod
    def from_dict(cls, data, schedules, **kwargs):
        """Create a DomesticHotWaterSetting from a dictionary.

        Args:
            data (dict): The python dictionary.
            schedules (dict): A dictionary of UmiSchedules with their id as keys.
            **kwargs: keywords passed the MaterialBase constructor.
        """
        _id = data.pop("$id")
        wat_sch = data.pop("WaterSchedule", None)
        schedule = schedules[wat_sch["$ref"]]
        return cls(id=_id, WaterSchedule=schedule, **data, **kwargs)

    def to_dict(self):
        """Return DomesticHotWaterSetting dictionary representation."""
        self.validate()  # Validate object before trying to get json format

        data_dict = collections.OrderedDict()

        data_dict["$id"] = str(self.id)
        data_dict["FlowRatePerFloorArea"] = round(self.FlowRatePerFloorArea, sigfigs=4)
        data_dict["IsOn"] = self.IsOn
        data_dict["WaterSchedule"] = self.WaterSchedule.to_ref()
        data_dict["WaterSupplyTemperature"] = round(
            self.WaterSupplyTemperature, sigfigs=4
        )
        data_dict["WaterTemperatureInlet"] = round(
            self.WaterTemperatureInlet, sigfigs=4
        )
        data_dict["Category"] = self.Category
        data_dict["Comments"] = validators.string(self.Comments, allow_empty=True)
        data_dict["DataSource"] = self.DataSource
        data_dict["Name"] = self.Name

        return data_dict

    @classmethod
    @timeit
    def from_zone(cls, zone_epbunch, **kwargs):
        """Create object from a zone EpBunch.

        WaterUse:Equipment objects referring to this zone will be parsed.

        Args:
            zone (EpBunch): The zone object.
        """
        # If Zone is not part of Conditioned Area, it should not have a DHW object.
        if zone_epbunch.Part_of_Total_Floor_Area.lower() == "no":
            return None

        # First, find the WaterUse:Equipment assigned to this zone
        dhw_objs = zone_epbunch.getreferingobjs(
            iddgroups=["Water Systems"], fields=["Zone_Name"]
        )
        if not dhw_objs:
            # Sometimes, some of the WaterUse:Equipment objects are not assigned to
            # any zone. Therefore, to account for their water usage, we can try to
            # assign dangling WaterUse:Equipments by looking for the zone name in the
            # object name.
            dhw_objs.extend(
                [
                    dhw
                    for dhw in zone_epbunch.theidf.idfobjects["WATERUSE:EQUIPMENT"]
                    if zone_epbunch.Name.lower() in dhw.Name.lower()
                ]
            )

        if dhw_objs:
            # This zone has more than one WaterUse:Equipment object
            zone_area = modeleditor.zonearea(zone_epbunch.theidf, zone_epbunch.Name)
            total_flow_rate = cls._do_flow_rate(dhw_objs, zone_area)
            water_schedule = cls._do_water_schedule(dhw_objs)
            inlet_temp = cls._do_inlet_temp(dhw_objs)
            supply_temp = cls._do_hot_temp(dhw_objs)

            name = zone_epbunch.Name + "_DHW"
            z_dhw = cls(
                Name=name,
                FlowRatePerFloorArea=total_flow_rate,
                IsOn=bool(total_flow_rate > 0),
                WaterSchedule=water_schedule,
                WaterSupplyTemperature=supply_temp,
                WaterTemperatureInlet=inlet_temp,
                Category=zone_epbunch.theidf.name,
                area=zone_area,
                **kwargs,
            )
            return z_dhw
        else:
            log(f"No 'Water Systems' found in zone '{zone_epbunch.Name}'")
            return None

    @classmethod
    def _do_hot_temp(cls, dhw_objs):
        """Resolve hot water temperature.

        Args:
            dhw_objs:
        """
        hot_schds = []
        for obj in dhw_objs:
            # Reference to the schedule object specifying the target water
            # temperature [C]. If blank, the target temperature defaults to
            # the hot water supply temperature.
            schedule_name = (
                obj.Target_Temperature_Schedule_Name
                if obj.Target_Temperature_Schedule_Name != ""
                else obj.Hot_Water_Supply_Temperature_Schedule_Name
            )
            epbunch = obj.theidf.schedules_dict[schedule_name.upper()]
            hot_schd = UmiSchedule.from_epbunch(epbunch)
            hot_schds.append(hot_schd)

        return np.array([sched.all_values.mean() for sched in hot_schds]).mean()

    @classmethod
    def _do_inlet_temp(cls, dhw_objs):
        """Calculate inlet water temperature."""
        WaterTemperatureInlet = []
        for obj in dhw_objs:
            if obj.Cold_Water_Supply_Temperature_Schedule_Name != "":
                # If a cold water supply schedule is provided, create the
                # schedule
                epbunch = obj.theidf.schedules_dict[
                    obj.Cold_Water_Supply_Temperature_Schedule_Name.upper()
                ]
                cold_schd_names = UmiSchedule.from_epbunch(epbunch)
                WaterTemperatureInlet.append(cold_schd_names.mean)
            else:
                # If blank, water temperatures are calculated by the
                # Site:WaterMainsTemperature object.
                water_mains_temps = obj.theidf.idfobjects[
                    "Site:WaterMainsTemperature".upper()
                ]
                if water_mains_temps:
                    # If a "Site:WaterMainsTemperature" object exists,
                    # do water depending on calc method:
                    water_mains_temp = water_mains_temps[0]
                    if water_mains_temp.Calculation_Method.lower() == "schedule":
                        # From Schedule method
                        mains_scd = UmiSchedule.from_epbunch(
                            obj.theidf.schedules_dict[
                                water_mains_temp.Schedule_Name.upper()
                            ]
                        )
                        WaterTemperatureInlet.append(mains_scd.mean())
                    elif water_mains_temp.Calculation_Method.lower() == "correlation":
                        # From Correlation method
                        mean_outair_temp = (
                            water_mains_temp.Annual_Average_Outdoor_Air_Temperature
                        )
                        max_dif = (
                            water_mains_temp.Maximum_Difference_In_Monthly_Average_Outdoor_Air_Temperatures
                        )

                        WaterTemperatureInlet.append(
                            water_main_correlation(mean_outair_temp, max_dif).mean()
                        )
                else:
                    # Else, there is no Site:WaterMainsTemperature object in
                    # the input file, a default constant value of 10 C is
                    # assumed.
                    WaterTemperatureInlet.append(float(10))
        return mean(WaterTemperatureInlet) if WaterTemperatureInlet else 10

    @classmethod
    def _do_water_schedule(cls, dhw_objs):
        """Return the WaterSchedule for a list of WaterUse:Equipment objects.

        If more than one objects are passed, a combined schedule is returned

        Args:
            dhw_objs (list of EpBunch): List of WaterUse:Equipment objects.

        Returns:
            UmiSchedule: The WaterSchedule
        """
        water_schds = [
            UmiSchedule.from_epbunch(
                obj.theidf.schedules_dict[obj.Flow_Rate_Fraction_Schedule_Name.upper()],
                quantity=obj.Peak_Flow_Rate,
            )
            for obj in dhw_objs
        ]

        return reduce(
            UmiSchedule.combine,
            water_schds,
            weights=None,
            quantity=True,
        )

    @classmethod
    def _do_flow_rate(cls, dhw_objs, area):
        """Calculate total flow rate from list of WaterUse:Equipment objects.

        The zone's net_conditioned_building_area property is used to normalize the
        flow rate.

        Args:
            dhw_objs (Idf_MSequence):
        """
        total_flow_rate = 0
        for obj in dhw_objs:
            total_flow_rate += obj.Peak_Flow_Rate  # m3/s
        total_flow_rate /= area  # m3/s/m2
        total_flow_rate *= 3600.0  # m3/h/m2
        return total_flow_rate

    def combine(self, other, **kwargs):
        """Combine two DomesticHotWaterSetting objects together.

        Notes:
            When combining 2 DomesticHotWater Settings objects, the WaterSchedule
            must be averaged via the final quantity which is the peak floor rate
            [m3/hr/m2] * area [m2].

        .. code-block:: python

            (
                np.average(
                    [zone_1.WaterSchedule.all_values, zone_2.WaterSchedule.all_values],
                    axis=0,
                    weights=[
                        zone_1.FlowRatePerFloorArea * zone_1.area,
                        zone_2.FlowRatePerFloorArea * zone_2.area,
                    ],
                )
                * (combined.FlowRatePerFloorArea * 100)
            ).sum()

        Args:
            other (DomesticHotWaterSetting): The other object.
            **kwargs: keywords passed to the constructor.

        Returns:
            (DomesticHotWaterSetting): a new combined object.
        """
        # Check if other is None. Simply return self
        if not other:
            return self

        if not self:
            return other

        # Check if other is the same type as self
        if not isinstance(other, self.__class__):
            msg = "Cannot combine %s with %s" % (
                self.__class__.__name__,
                other.__class__.__name__,
            )
            raise NotImplementedError(msg)

        # Check if other is not the same as self
        if self == other:
            return self

        meta = self._get_predecessors_meta(other)
        new_obj = DomesticHotWaterSetting(
            WaterSchedule=UmiSchedule.combine(
                self.WaterSchedule,
                other.WaterSchedule,
                weights=[
                    self.FlowRatePerFloorArea * self.area,
                    other.FlowRatePerFloorArea * other.area,
                ],
            ),
            IsOn=any((self.IsOn, other.IsOn)),
            FlowRatePerFloorArea=self.float_mean(
                other, "FlowRatePerFloorArea", [self.area, other.area]
            ),
            WaterSupplyTemperature=self.float_mean(
                other, "WaterSupplyTemperature", [self.area, other.area]
            ),
            WaterTemperatureInlet=self.float_mean(
                other, "WaterTemperatureInlet", [self.area, other.area]
            ),
            area=self.area + other.area,
            **meta,
        )
        new_obj.predecessors.update(self.predecessors + other.predecessors)
        return new_obj

    def validate(self):
        """Validate object and fill in missing values."""
        return self

    @classmethod
    def whole_building(cls, idf):
        """Create one DomesticHotWaterSetting for whole building model.

        Args:
            idf (IDF): The idf model.

        Returns:
            DomesticHotWaterSetting: The DomesticHotWaterSetting object.
        """
        # Unconditioned area could be zero, therefore taking max of both
        area = max(idf.net_conditioned_building_area, idf.unconditioned_building_area)

        z_dhw_list = []
        dhw_objs = idf.idfobjects["WaterUse:Equipment".upper()]
        if not dhw_objs:
            # defaults with 0 flow rate.
            total_flow_rate = 0
            water_schedule = UmiSchedule.constant_schedule()
            supply_temp = 60
            inlet_temp = 10

            name = idf.name + "_DHW"
            z_dhw = DomesticHotWaterSetting(
                WaterSchedule=water_schedule,
                IsOn=bool(total_flow_rate > 0),
                FlowRatePerFloorArea=total_flow_rate,
                WaterSupplyTemperature=supply_temp,
                WaterTemperatureInlet=inlet_temp,
                area=area,
                Name=name,
                Category=idf.name,
            )
            z_dhw_list.append(z_dhw)
        else:
            total_flow_rate = DomesticHotWaterSetting._do_flow_rate(dhw_objs, area)
            water_schedule = DomesticHotWaterSetting._do_water_schedule(dhw_objs)
            water_schedule.quantity = total_flow_rate
            inlet_temp = DomesticHotWaterSetting._do_inlet_temp(dhw_objs)
            supply_temp = DomesticHotWaterSetting._do_hot_temp(dhw_objs)
            z_dhw = DomesticHotWaterSetting(
                WaterSchedule=water_schedule,
                IsOn=bool(total_flow_rate > 0),
                FlowRatePerFloorArea=total_flow_rate,
                WaterSupplyTemperature=supply_temp,
                WaterTemperatureInlet=inlet_temp,
                area=area,
                Name="Whole Building WaterUse:Equipment",
                Category=idf.name,
            )
            z_dhw_list.append(z_dhw)

        return reduce(DomesticHotWaterSetting.combine, z_dhw_list)

    def mapping(self, validate=True):
        """Get a dict based on the object properties, useful for dict repr.

        Args:
            validate (bool): If True, try to validate object before returning the
                mapping.
        """
        if validate:
            self.validate()

        return dict(
            FlowRatePerFloorArea=self.FlowRatePerFloorArea,
            IsOn=self.IsOn,
            WaterSchedule=self.WaterSchedule,
            WaterSupplyTemperature=self.WaterSupplyTemperature,
            WaterTemperatureInlet=self.WaterTemperatureInlet,
            Category=self.Category,
            Comments=self.Comments,
            DataSource=self.DataSource,
            Name=self.Name,
        )

    def duplicate(self):
        """Get copy of self."""
        return self.__copy__()

    def __add__(self, other):
        """Overload + to implement self.combine.

        Args:
            other (DomesticHotWaterSetting):
        """
        return self.combine(
            other,
        )

    def __hash__(self):
        """Return the hash value of self."""
        return hash(
            (self.__class__.__name__, getattr(self, "Name", None), self.DataSource)
        )

    def __key__(self):
        """Get a tuple of attributes. Useful for hashing and comparing."""
        return (
            self.IsOn,
            self.FlowRatePerFloorArea,
            self.WaterSupplyTemperature,
            self.WaterTemperatureInlet,
            self.WaterSchedule,
        )

    def __eq__(self, other):
        """Assert self is equivalent to other."""
        if not isinstance(other, DomesticHotWaterSetting):
            return NotImplemented
        else:
            return self.__key__() == other.__key__()

    def __str__(self):
        """Return string representation."""
        return (
            f"{str(self.id)}: {str(self.Name)} "
            f"PeakFlow {self.FlowRatePerFloorArea:.5f} m3/hr/m2"
        )

    def __copy__(self):
        """Create a copy of self."""
        return self.__class__(**self.mapping(validate=False))


def water_main_correlation(t_out_avg, max_diff):
    """Based on the correlation developed by Craig Christensen and Jay Burch.

    Returns a 365 days temperature profile.

    Info:
        https://bigladdersoftware.com/epx/docs/8-9/engineering-reference
        /water-systems.html#water-mains-temperatures

    Args:
        t_out_avg (float): average annual outdoor air temperature (°C).
        max_diff (float): maximum difference in monthly average outdoor air
            temperatures (°C).

    Returns:
        (pd.Series): water mains temperature profile.
    """
    import numpy as np
    import pandas as pd

    Q_ = settings.unit_registry.Quantity
    t_out_avg_F = Q_(t_out_avg, "degC").to("degF")
    max_diff_F = Q_(max_diff, "delta_degC").to("delta_degF")
    ratio = 0.4 + 0.01 * (t_out_avg_F.m - 44)
    lag = 35 - 1.0 * (t_out_avg_F.m - 44)
    days = np.arange(1, 365)

    def function(t_out_avg, day, max_diff):
        return (t_out_avg + 6) + ratio * (max_diff / 2) * np.sin(
            np.deg2rad(0.986 * (day - 15 - lag) - 90)
        )

    mains = [Q_(function(t_out_avg_F.m, day, max_diff_F.m), "degF") for day in days]
    series = pd.Series([temp.to("degC").m for temp in mains])
    return series



"""EnergyPlus variables module."""

import pandas as pd
from energy_pandas import EnergyDataFrame
from geomeppy.patches import EpBunch

from archetypal.idfclass.extensions import bunch2db
from archetypal.reportdata import ReportData


class Variable:
    """Variable class.

    Holds values for a specific Variable.
    """

    def __init__(self, idf, variable: (dict or EpBunch)):
        """Initialize a Meter object."""
        self._idf = idf
        if isinstance(variable, dict):
            self._key = variable.pop("key").upper()
            self._epobject = self._idf.anidfobject(key=self._key, **variable)
        elif isinstance(variable, EpBunch):
            self._key = variable.key
            self._epobject = variable
        else:
            raise TypeError()

    def __repr__(self):
        """Return a representation of self."""
        return self._epobject.__str__()

    def values(
        self,
        units=None,
        reporting_frequency="Hourly",
        normalize=False,
        sort_values=False,
    ):
        """Return the Variable as a time-series (:class:`EnergySeries`).

        Can return an :class:`EnergyDataFrame` with multiple columns.

        Specify the reporting_frequency such as "Hourly" or "Monthly".
        It is possible to convert the time-series to another unit, e.g.: "J" to "kWh".

        Data is retrieved from the sql file.

        Args:
            units (str): Convert original values to another unit. The original unit
                is detected automatically and a dimensionality check is performed.
            reporting_frequency (str): Timestep, Hourly, Daily, Monthly,
                RunPeriod, Environment, Annual or Detailed. Default "Hourly".
            normalize (bool): Normalize between 0 and 1.
            sort_values (bool): If True, values are sorted (default ascending=True).

        Returns:
            EnergyDataFrame: The time-series object.
        """
        self._epobject.Reporting_Frequency = reporting_frequency.lower()
        if self._epobject not in self._idf.idfobjects[self._epobject.key]:
            self._idf.addidfobject(self._epobject)
            self._idf.simulate()
        report = ReportData.from_sqlite(
            sqlite_file=self._idf.sql_file,
            table_name=self._epobject.Variable_Name,
            environment_type=1 if self._idf.design_day else 3,
            reporting_frequency=bunch2db[reporting_frequency],
        )
        return EnergyDataFrame.from_reportdata(
            report,
            name=self._epobject.Variable_Name,
            normalize=normalize,
            sort_values=sort_values,
            to_units=units,
        )


class VariableGroup:
    """A class for sub variable groups (Output:Variable)."""

    def __init__(self, idf, variables_dict: dict):
        """Initialize VariableGroup."""
        self._idf = idf
        self._properties = {}

        for i, variable in variables_dict.items():
            variable_name = (
                variable["Variable_Name"].replace(":", "__").replace(" ", "_")
            )
            self._properties[variable_name] = Variable(idf, variable)
            setattr(self, variable_name, self._properties[variable_name])

    def __getitem__(self, variable_name):
        """Get item by key."""
        return self._properties[variable_name]


class Variables:
    """Lists available variables in the IDF model.

    Once simulated at least once, he IDF.variables attribute is populated with
    variable categories  and each category is populated with all the available
    variables.

    Example:
        For example, to retrieve the "Zone Operative Temperature" variable, simply call

        .. code-block::

            >>> from archetypal import IDF
            >>> idf = IDF()  # load an actual idf file here
            >>> idf.variables.OutputVariable.Zone_Operative_Temperature.values()

    Hint:
        Available meters are read from the .mdd file
    """

    OutputVariable = VariableGroup  # Placeholder for variables

    def __init__(self, idf):
        """Initialize MeterGroup."""
        self._idf = idf

        rdd, *_ = self._idf.simulation_dir.files("*.rdd")

        if not rdd:
            raise FileNotFoundError
        variables = pd.read_csv(
            rdd,
            skiprows=2,
            names=["key", "Key_Value", "Variable_Name", "Reporting_Frequency"],
        )
        variables.Reporting_Frequency = variables.Reporting_Frequency.str.replace(
            r"\;.*", ""
        )
        for key, group in variables.groupby("key"):
            variable_dict = group.T.to_dict()
            setattr(
                self,
                key.replace(":", "").replace(" ", "_"),
                VariableGroup(self._idf, variable_dict),
            )



"""archetypal ZoneDefinition module."""

import collections
import sqlite3
import time

from eppy.bunch_subclass import BadEPFieldError
from sigfig import round
from validator_collection import validators

from archetypal.template.conditioning import ZoneConditioning
from archetypal.template.constructions.internal_mass import InternalMass
from archetypal.template.constructions.opaque_construction import OpaqueConstruction
from archetypal.template.dhw import DomesticHotWaterSetting
from archetypal.template.load import ZoneLoad
from archetypal.template.umi_base import UmiBase
from archetypal.template.ventilation import VentilationSetting
from archetypal.template.window_setting import WindowSetting
from archetypal.template.zone_construction_set import ZoneConstructionSet
from archetypal.utils import log, settings


class ZoneDefinition(UmiBase):
    """Zone settings class.

    .. image:: ../images/template/zoneinfo-zone.png
    """

    __slots__ = (
        "_internal_mass_exposed_per_floor_area",
        "_constructions",
        "_loads",
        "_conditioning",
        "_ventilation",
        "_domestic_hot_water",
        "_windows",
        "_occupants",
        "_daylight_mesh_resolution",
        "_daylight_workplane_height",
        "_internal_mass_construction",
        "_is_part_of_conditioned_floor_area",
        "_is_part_of_total_floor_area",
        "_zone_surfaces",
        "_volume",
        "_multiplier",
        "_area",
        "_is_core",
    )

    def __init__(
        self,
        Name,
        Constructions=None,
        Loads=None,
        Conditioning=None,
        Ventilation=None,
        DomesticHotWater=None,
        DaylightMeshResolution=1,
        DaylightWorkplaneHeight=0.8,
        InternalMassConstruction=None,
        InternalMassExposedPerFloorArea=1.05,
        Windows=None,
        area=1,
        volume=1,
        occupants=1,
        is_part_of_conditioned_floor_area=True,
        is_part_of_total_floor_area=True,
        multiplier=1,
        zone_surfaces=None,
        is_core=False,
        **kwargs,
    ):
        """Initialize :class:`Zone` object.

        Args:
            Name (str): Name of the object. Must be Unique.
            Constructions (ZoneConstructionSet):
            Loads (ZoneLoad): Loads of the zone defined with the lights,
                equipment and occupancy parameters (see :class:`ZoneLoad`)
            Conditioning (ZoneConditioning): Conditioning of the zone defined
                with heating/cooling and mechanical ventilation parameters (see
                :class:`ZoneConditioning`)
            Ventilation (VentilationSetting): Ventilation settings of the zone
                defined with the infiltration rate and natural ventilation
                parameters (see :class:`VentilationSetting`)
            DomesticHotWater (archetypal.template.dhw.DomesticHotWaterSetting):
            DaylightMeshResolution (float):
            DaylightWorkplaneHeight (float):
            InternalMassConstruction (archetypal.OpaqueConstruction):
            InternalMassExposedPerFloorArea:
            Windows (WindowSetting): The WindowSetting object associated with
                this zone.
            area (float):
            volume (float):
            occupants (float):
            **kwargs:
        """
        super(ZoneDefinition, self).__init__(Name, **kwargs)

        self.Ventilation = Ventilation
        self.Loads = Loads
        self.Conditioning = Conditioning
        self.Constructions = Constructions
        self.DaylightMeshResolution = DaylightMeshResolution
        self.DaylightWorkplaneHeight = DaylightWorkplaneHeight
        self.DomesticHotWater = DomesticHotWater
        self.InternalMassConstruction = InternalMassConstruction
        self.InternalMassExposedPerFloorArea = InternalMassExposedPerFloorArea

        self.Windows = Windows  # This is not used in to_dict()

        if zone_surfaces is None:
            zone_surfaces = []
        self.zone_surfaces = zone_surfaces
        self.area = area
        self.volume = volume
        self.occupants = occupants
        self.is_part_of_conditioned_floor_area = is_part_of_conditioned_floor_area
        self.is_part_of_total_floor_area = is_part_of_total_floor_area
        self.multiplier = multiplier
        self.is_core = is_core

    @property
    def Constructions(self):
        """Get or set the ZoneConstructionSet object."""
        return self._constructions

    @Constructions.setter
    def Constructions(self, value):
        if value is not None:
            assert isinstance(value, ZoneConstructionSet), (
                f"Input value error. Constructions must be of "
                f"type {ZoneConstructionSet}, not {type(value)}."
            )
        self._constructions = value

    @property
    def Loads(self):
        """Get or set the ZoneLoad object."""
        return self._loads

    @Loads.setter
    def Loads(self, value):
        if value is not None:
            assert isinstance(value, ZoneLoad), (
                f"Input value error. Loads must be of "
                f"type {ZoneLoad}, not {type(value)}."
            )
        self._loads = value

    @property
    def Conditioning(self):
        """Get or set the ZoneConditioning object."""
        return self._conditioning

    @Conditioning.setter
    def Conditioning(self, value):
        if value is not None:
            assert isinstance(value, ZoneConditioning), (
                f"Input value error. Conditioning must be of "
                f"type {ZoneConditioning}, not {type(value)}."
            )
        self._conditioning = value

    @property
    def Ventilation(self):
        """Get or set the VentilationSetting object."""
        return self._ventilation

    @Ventilation.setter
    def Ventilation(self, value):
        if value is not None:
            assert isinstance(value, VentilationSetting), (
                f"Input value error. Ventilation must be of "
                f"type {VentilationSetting}, not {type(value)}."
            )
        self._ventilation = value

    @property
    def DomesticHotWater(self):
        """Get or set the DomesticHotWaterSetting object."""
        return self._domestic_hot_water

    @DomesticHotWater.setter
    def DomesticHotWater(self, value):
        if value is not None:
            assert isinstance(value, DomesticHotWaterSetting), (
                f"Input value error. DomesticHotWater must be of "
                f"type {DomesticHotWaterSetting}, not {type(value)}."
            )
        self._domestic_hot_water = value

    @property
    def DaylightMeshResolution(self):
        """Get or set the daylight mesh resolution [m]."""
        return self._daylight_mesh_resolution

    @DaylightMeshResolution.setter
    def DaylightMeshResolution(self, value):
        self._daylight_mesh_resolution = validators.float(value, minimum=0)

    @property
    def DaylightWorkplaneHeight(self):
        """Get or set the DaylightWorkplaneHeight [m]."""
        return self._daylight_workplane_height

    @DaylightWorkplaneHeight.setter
    def DaylightWorkplaneHeight(self, value):
        self._daylight_workplane_height = validators.float(value, minimum=0)

    @property
    def InternalMassConstruction(self):
        """Get or set the internal mass construction object."""
        return self._internal_mass_construction

    @InternalMassConstruction.setter
    def InternalMassConstruction(self, value):
        if value is not None:
            assert isinstance(value, OpaqueConstruction), (
                f"Input value error. InternalMassConstruction must be of "
                f"type {OpaqueConstruction}, not {type(value)}."
            )
        self._internal_mass_construction = value

    @property
    def InternalMassExposedPerFloorArea(self):
        """Get or set the internal mass exposed per floor area [-]."""
        return self._internal_mass_exposed_per_floor_area

    @InternalMassExposedPerFloorArea.setter
    def InternalMassExposedPerFloorArea(self, value):
        self._internal_mass_exposed_per_floor_area = validators.float(value, minimum=0)

    @property
    def Windows(self):
        """Get or set the WindowSetting object."""
        return self._windows

    @Windows.setter
    def Windows(self, value):
        if value is not None:
            assert isinstance(value, WindowSetting), (
                f"Input value error. Windows must be of "
                f"type {WindowSetting}, not {type(value)}."
            )
        self._windows = value

    @property
    def occupants(self):
        """Get or set the number of occupants in the zone."""
        return self._occupants

    @occupants.setter
    def occupants(self, value):
        self._occupants = value

    @property
    def area(self):
        """Get or set the area of the zone [m²]."""
        return self._area

    @area.setter
    def area(self, value):
        self._area = validators.float(value, minimum=0)

    @property
    def volume(self):
        """Get or set the volume of the zone [m³]."""
        return self._volume

    @volume.setter
    def volume(self, value):
        self._volume = validators.float(value, minimum=0)

    @property
    def is_core(self):
        """Get or set if the zone is a core zone [bool]."""
        return self._is_core

    @is_core.setter
    def is_core(self, value):
        assert isinstance(value, bool), value
        self._is_core = value

    @property
    def multiplier(self):
        """Get or set the zone multiplier.

        Note: Zone multiplier is designed as a “multiplier” for floor
        area, zone loads, and energy consumed by internal gains.
        """
        return self._multiplier

    @multiplier.setter
    def multiplier(self, value):
        self._multiplier = validators.integer(value, minimum=1)

    @property
    def is_part_of_conditioned_floor_area(self):
        """Get or set is part of conditioned area [bool]."""
        return self._is_part_of_conditioned_floor_area

    @is_part_of_conditioned_floor_area.setter
    def is_part_of_conditioned_floor_area(self, value):
        assert isinstance(value, bool)
        self._is_part_of_conditioned_floor_area = value

    @property
    def is_part_of_total_floor_area(self):
        """Get or set is part od the total building floor area [bool]."""
        return self._is_part_of_total_floor_area

    @is_part_of_total_floor_area.setter
    def is_part_of_total_floor_area(self, value):
        assert isinstance(value, bool)
        self._is_part_of_total_floor_area = value

    @property
    def zone_surfaces(self):
        """Get or set the list of surfaces for this zone."""
        return self._zone_surfaces

    @zone_surfaces.setter
    def zone_surfaces(self, value):
        self._zone_surfaces = validators.iterable(value, allow_empty=True)

    def to_dict(self):
        """Return ZoneDefinition dictionary representation."""
        self.validate()  # Validate object before trying to get json format

        data_dict = collections.OrderedDict()

        data_dict["$id"] = str(self.id)
        data_dict["Conditioning"] = self.Conditioning.to_ref()
        data_dict["Constructions"] = self.Constructions.to_ref()
        data_dict["DaylightMeshResolution"] = round(self.DaylightMeshResolution, 2)
        data_dict["DaylightWorkplaneHeight"] = round(self.DaylightWorkplaneHeight, 2)
        data_dict["DomesticHotWater"] = self.DomesticHotWater.to_ref()
        data_dict["InternalMassConstruction"] = self.InternalMassConstruction.to_ref()
        data_dict["InternalMassExposedPerFloorArea"] = round(
            self.InternalMassExposedPerFloorArea, 3
        )
        data_dict["Loads"] = self.Loads.to_ref()
        data_dict["Ventilation"] = self.Ventilation.to_ref()
        data_dict["Category"] = self.Category
        data_dict["Comments"] = validators.string(self.Comments, allow_empty=True)
        data_dict["DataSource"] = self.DataSource
        data_dict["Name"] = self.Name

        return data_dict

    @classmethod
    def from_dict(
        cls,
        data,
        zone_conditionings,
        zone_construction_sets,
        domestic_hot_water_settings,
        opaque_constructions,
        zone_loads,
        ventilation_settings,
        **kwargs,
    ):
        """Create a ZoneDefinition from a dictionary.

        Args:
            data (dict): The python dictionary.
            zone_conditionings (dict): A dictionary of ZoneConditioning objects with
                their id as keys.
            zone_construction_sets (dict): A dictionary of ZoneConstructionSet
                objects with their id as keys.
            domestic_hot_water_settings (dict): A dictionary of DomesticHotWaterSetting
                objects with their id as keys.
            opaque_constructions (dict): A dictionary of OpaqueConstruction objects
                with their id as keys.
            zone_loads (dict): A dictionary of ZoneLoad objects with their id as
                keys.
            ventilation_settings (dict): A dictionary of ZoneConditioning objects with
                their id as keys.
            **kwargs: keywords passed to the constructor.

        .. code-block:: python

            {
              "$id": "175",
              "Conditioning": {
                "$ref": "165"
              },
              "Constructions": {
                "$ref": "168"
              },
              "DaylightMeshResolution": 1.0,
              "DaylightWorkplaneHeight": 0.8,
              "DomesticHotWater": {
                "$ref": "159"
              },
              "InternalMassConstruction": {
                "$ref": "54"
              },
              "InternalMassExposedPerFloorArea": 1.05,
              "Loads": {
                "$ref": "172"
              },
              "Ventilation": {
                "$ref": "162"
              },
              "Category": "Office Spaces",
              "Comments": null,
              "DataSource": "MIT_SDL",
              "Name": "B_Off_0"
            }
        """
        _id = data.pop("$id")

        conditioning = zone_conditionings[data.pop("Conditioning")["$ref"]]
        construction_set = zone_construction_sets[data.pop("Constructions")["$ref"]]
        domestic_hot_water_setting = domestic_hot_water_settings[
            data.pop("DomesticHotWater")["$ref"]
        ]
        internal_mass_construction = opaque_constructions[
            data.pop("InternalMassConstruction")["$ref"]
        ]
        zone_load = zone_loads[data.pop("Loads")["$ref"]]
        ventilation_setting = ventilation_settings[data.pop("Ventilation")["$ref"]]

        return cls(
            id=_id,
            Conditioning=conditioning,
            Constructions=construction_set,
            DomesticHotWater=domestic_hot_water_setting,
            InternalMassConstruction=internal_mass_construction,
            Loads=zone_load,
            Ventilation=ventilation_setting,
            **data,
            **kwargs,
        )

    @classmethod
    def from_epbunch(cls, ep_bunch, construct_parents=True, **kwargs):
        """Create a Zone object from an eppy 'ZONE' epbunch.

        Args:
            ep_bunch (eppy.bunch_subclass.EpBunch): The Zone EpBunch.
            construct_parents (bool): If False, skips construction of parents objects
                such as Constructions, Conditioning, etc.
        """
        assert (
            ep_bunch.key.lower() == "zone"
        ), f"Expected a `ZONE` epbunch, got {ep_bunch.key}"
        start_time = time.time()
        log('Constructing :class:`Zone` for zone "{}"'.format(ep_bunch.Name))

        def calc_zone_area(zone_ep):
            """Get zone area from simulation sql file."""
            with sqlite3.connect(zone_ep.theidf.sql_file) as conn:
                sql_query = """
                    SELECT t.Value 
                    FROM TabularDataWithStrings t 
                    WHERE TableName='Zone Summary' and ColumnName='Area' and RowName=?
                """
                (res,) = conn.execute(sql_query, (zone_ep.Name.upper(),)).fetchone()
            return float(res)

        def calc_zone_volume(zone_ep):
            """Get zone volume from simulation sql file."""
            with sqlite3.connect(zone_ep.theidf.sql_file) as conn:
                sql_query = (
                    "SELECT t.Value FROM TabularDataWithStrings t "
                    "WHERE TableName='Zone Summary' and ColumnName='Volume' and "
                    "RowName=?"
                )
                (res,) = conn.execute(sql_query, (zone_ep.Name.upper(),)).fetchone()
            return float(res)

        def calc_zone_occupants(zone_ep):
            """Get zone occupants from simulation sql file."""
            with sqlite3.connect(zone_ep.theidf.sql_file) as conn:
                sql_query = (
                    "SELECT t.Value FROM TabularDataWithStrings t "
                    "WHERE TableName='Average Outdoor Air During Occupied Hours' and ColumnName='Nominal Number of Occupants' and RowName=?"
                )

                fetchone = conn.execute(sql_query, (zone_ep.Name.upper(),)).fetchone()
                (res,) = fetchone or (0,)
            return float(res)

        def calc_is_part_of_conditioned_floor_area(zone_ep):
            """Return True if zone is part of the conditioned floor area."""
            with sqlite3.connect(zone_ep.theidf.sql_file) as conn:
                sql_query = (
                    "SELECT t.Value FROM TabularDataWithStrings t WHERE "
                    "TableName='Zone Summary' and ColumnName='Conditioned (Y/N)' "
                    "and RowName=?"
                    ""
                )
                res = conn.execute(sql_query, (zone_ep.Name.upper(),)).fetchone()
            return "Yes" in res

        def calc_is_part_of_total_floor_area(zone_ep):
            """Return True if zone is part of the total floor area."""
            with sqlite3.connect(zone_ep.theidf.sql_file) as conn:
                sql_query = (
                    "SELECT t.Value FROM TabularDataWithStrings t WHERE "
                    "TableName='Zone Summary' and ColumnName='Part of "
                    "Total Floor Area (Y/N)' and RowName=?"
                )
                res = conn.execute(sql_query, (zone_ep.Name.upper(),)).fetchone()
            return "Yes" in res

        def calc_multiplier(zone_ep):
            """Get the zone multiplier from simulation sql."""
            with sqlite3.connect(zone_ep.theidf.sql_file) as conn:
                sql_query = (
                    "SELECT t.Value FROM TabularDataWithStrings t WHERE "
                    "TableName='Zone Summary' and "
                    "ColumnName='Multipliers' and RowName=?"
                )
                (res,) = conn.execute(sql_query, (zone_ep.Name.upper(),)).fetchone()
            return int(float(res))

        def is_core(zone_ep):
            # if all surfaces don't have boundary condition == "Outdoors"
            iscore = True
            for s in zone_ep.zonesurfaces:
                try:
                    if (abs(int(s.tilt)) < 180) & (abs(int(s.tilt)) > 0):
                        obc = s.Outside_Boundary_Condition.lower()
                        if obc in ["outdoors", "ground"]:
                            iscore = False
                            break
                except BadEPFieldError:
                    pass  # pass surfaces that don't have an OBC,
                    # eg. InternalMass
            return iscore

        name = ep_bunch.Name
        zone = cls(
            Name=name,
            Category=ep_bunch.theidf.name,
            area=calc_zone_area(ep_bunch),
            volume=calc_zone_volume(ep_bunch),
            occupants=calc_zone_occupants(ep_bunch),
            is_part_of_conditioned_floor_area=calc_is_part_of_conditioned_floor_area(
                ep_bunch
            ),
            is_part_of_total_floor_area=calc_is_part_of_total_floor_area(ep_bunch),
            multiplier=calc_multiplier(ep_bunch),
            zone_surfaces=ep_bunch.zonesurfaces,
            is_core=is_core(ep_bunch),
            **kwargs,
        )

        if construct_parents:
            zone.Constructions = ZoneConstructionSet.from_zone(zone, **kwargs)
            zone.Conditioning = ZoneConditioning.from_zone(zone, ep_bunch, **kwargs)
            zone.Ventilation = VentilationSetting.from_zone(zone, ep_bunch, **kwargs)
            zone.DomesticHotWater = DomesticHotWaterSetting.from_zone(
                ep_bunch, **kwargs
            )
            zone.Loads = ZoneLoad.from_zone(zone, ep_bunch, **kwargs)
            internal_mass_from_zone = InternalMass.from_zone(ep_bunch)
            zone.InternalMassConstruction = internal_mass_from_zone.construction
            zone.InternalMassExposedPerFloorArea = (
                internal_mass_from_zone.total_area_exposed_to_zone
            )
            zone.Windows = WindowSetting.from_zone(zone, **kwargs)

        log(
            'completed Zone "{}" constructor in {:,.2f} seconds'.format(
                ep_bunch.Name, time.time() - start_time
            )
        )
        return zone

    def combine(self, other, weights=None, allow_duplicates=False):
        """Combine two ZoneDefinition objects together.

        Args:
            other (ZoneDefinition): The other object.
            weights (list-like, optional): A list-like object of len 2. If None,
                the volume of the zones for which self and other belongs is
                used.

        Todo:
            Create Equivalent InternalMassConstruction from partitions when combining
            zones.

        Returns:
            (ZoneDefinition): the combined Zone object.
        """
        # Check if other is None. Simply return self
        if not self and not other:
            return None
        elif self == other:
            area = 1 if self.area + other.area == 2 else self.area + other.area
            volume = (
                1 if self.volume + other.volume == 2 else self.volume + other.volume
            )
            new_obj = self.duplicate()
            new_obj.area = area
            new_obj.volume = volume
            return new_obj
        elif not self or not other:
            new_obj = (self or other).duplicate()
            return new_obj

        # Check if other is the same type as self
        if not isinstance(other, self.__class__):
            msg = "Cannot combine %s with %s" % (
                self.__class__.__name__,
                other.__class__.__name__,
            )
            raise NotImplementedError(msg)

        meta = self._get_predecessors_meta(other)

        if not weights:
            zone_weight = settings.zone_weight
            weights = [
                getattr(self, str(zone_weight)),
                getattr(other, str(zone_weight)),
            ]
            log(
                'using zone {} "{}" as weighting factor in "{}" '
                "combine.".format(
                    zone_weight,
                    " & ".join(list(map(str, map(int, weights)))),
                    self.__class__.__name__,
                )
            )

        new_attr = dict(
            Conditioning=ZoneConditioning.combine(
                self.Conditioning, other.Conditioning, weights
            ),
            Constructions=ZoneConstructionSet.combine(
                self.Constructions, other.Constructions, weights
            ),
            Ventilation=VentilationSetting.combine(self.Ventilation, other.Ventilation),
            Windows=WindowSetting.combine(self.Windows, other.Windows, weights),
            DaylightMeshResolution=self.float_mean(
                other, "DaylightMeshResolution", weights=weights
            ),
            DaylightWorkplaneHeight=self.float_mean(
                other, "DaylightWorkplaneHeight", weights
            ),
            DomesticHotWater=DomesticHotWaterSetting.combine(
                self.DomesticHotWater, other.DomesticHotWater
            ),
            InternalMassConstruction=OpaqueConstruction.combine(
                self.InternalMassConstruction, other.InternalMassConstruction
            ),
            InternalMassExposedPerFloorArea=self.float_mean(
                other, "InternalMassExposedPerFloorArea", weights
            ),
            Loads=ZoneLoad.combine(self.Loads, other.Loads, weights),
        )
        new_obj = ZoneDefinition(**meta, **new_attr)

        # transfer aggregated values [volume, area, occupants] to new combined zone
        new_obj.volume = self.volume + other.volume
        new_obj.area = self.area + other.area
        new_obj.occupants = self.occupants + other.occupants

        if new_attr["Windows"]:  # Could be None
            new_attr["Windows"].area = new_obj.area

        new_obj.predecessors.update(self.predecessors + other.predecessors)
        return new_obj

    def validate(self):
        """Validate object and fill in missing values."""
        if not self.InternalMassConstruction:
            internal_mass = InternalMass.generic_internalmass_from_zone(self)
            self.InternalMassConstruction = internal_mass.construction
            self.InternalMassExposedPerFloorArea = (
                internal_mass.total_area_exposed_to_zone
            )
            log(
                f"While validating {self}, the required attribute "
                f"'InternalMassConstruction' was filled "
                f"with {self.InternalMassConstruction} and the "
                f"'InternalMassExposedPerFloorArea' set to"
                f" {self.InternalMassExposedPerFloorArea}"
            )

        if self.Conditioning is None:
            self.Conditioning = ZoneConditioning(Name="Unconditioned Zone")

        return self

    def mapping(self, validate=True):
        """Get a dict based on the object properties, useful for dict repr.

        Args:
            validate (bool): If True, try to validate object before returning the
                mapping.
        """
        if validate:
            self.validate()

        return dict(
            Conditioning=self.Conditioning,
            Constructions=self.Constructions,
            DaylightMeshResolution=self.DaylightMeshResolution,
            DaylightWorkplaneHeight=self.DaylightWorkplaneHeight,
            DomesticHotWater=self.DomesticHotWater,
            InternalMassConstruction=self.InternalMassConstruction,
            InternalMassExposedPerFloorArea=self.InternalMassExposedPerFloorArea,
            Windows=self.Windows,
            Loads=self.Loads,
            Ventilation=self.Ventilation,
            Category=self.Category,
            Comments=self.Comments,
            DataSource=self.DataSource,
            Name=self.Name,
            area=self.area,
            volume=self.volume,
            occupants=self.occupants,
            is_part_of_conditioned_floor_area=self.is_part_of_conditioned_floor_area,
            is_part_of_total_floor_area=self.is_part_of_total_floor_area,
            multiplier=self.multiplier,
            zone_surfaces=self.zone_surfaces,
            is_core=self.is_core,
        )

    def __add__(self, other):
        """Return a combination of self and other."""
        return ZoneDefinition.combine(self, other)

    def __hash__(self):
        """Return the hash value of self."""
        return hash(
            (self.__class__.__name__, getattr(self, "Name", None), self.DataSource)
        )

    def __eq__(self, other):
        """Assert self is equivalent to other."""
        if not isinstance(other, ZoneDefinition):
            return NotImplemented
        else:
            return all(
                [
                    self.Conditioning == other.Conditioning,
                    self.Constructions == other.Constructions,
                    self.DomesticHotWater == other.DomesticHotWater,
                    self.Loads == other.Loads,
                    self.Ventilation == other.Ventilation,
                    self.Windows == other.Windows,
                    self.InternalMassConstruction == other.InternalMassConstruction,
                    self.InternalMassExposedPerFloorArea
                    == other.InternalMassExposedPerFloorArea,
                    self.DaylightMeshResolution == other.DaylightMeshResolution,
                    self.DaylightWorkplaneHeight == other.DaylightWorkplaneHeight,
                ]
            )

    def __copy__(self):
        """Return a copy of self."""
        return self.__class__(**self.mapping(validate=False))


def resolve_obco(ep_bunch):
    """Resolve the outside boundary condition of a surface.

    Args:
        ep_bunch (EpBunch): The surface for which we are identifying the boundary
            object.

    Returns:
        (EpBunch, EpBunch): A tuple of:

            EpBunch: The other surface EpBunch: The other zone

    Notes:
        Info on the Outside Boundary Condition Object of a surface of type
        BuildingSurface:Detailed:

        Non-blank only if the field `Outside Boundary Condition` is *Surface*,
        *Zone*, *OtherSideCoefficients* or *OtherSideConditionsModel*. If
        Surface, specify name of corresponding surface in adjacent zone or
        specify current surface name for internal partition separating like
        zones. If Zone, specify the name of the corresponding zone and the
        program will generate the corresponding interzone surface. If
        Foundation, specify the name of the corresponding Foundation object and
        the program will calculate the heat transfer appropriately. If
        OtherSideCoefficients, specify name of
        SurfaceProperty:OtherSideCoefficients. If OtherSideConditionsModel,
        specify name of SurfaceProperty:OtherSideConditionsModel.
    """
    obc = ep_bunch.Outside_Boundary_Condition

    if obc.upper() == "ZONE":
        name = ep_bunch.Outside_Boundary_Condition_Object
        adj_zone = ep_bunch.theidf.getobject("ZONE", name)
        return None, adj_zone

    elif obc.upper() == "SURFACE":
        obco = ep_bunch.get_referenced_object("Outside_Boundary_Condition_Object")
        adj_zone = obco.theidf.getobject("ZONE", obco.Zone_Name)
        return obco, adj_zone
    else:
        return None, None


def is_core(zone):
    """Return true if zone is a core zone.

    Args:
        zone (eppy.bunch_subclass.EpBunch): The Zone object.

    Returns:
        (bool): Whether the zone is a core zone or not.
    """
    # if all surfaces don't have boundary condition == "Outdoors"
    iscore = True
    for s in zone.zonesurfaces:
        try:
            if (abs(int(s.tilt)) < 180) & (abs(int(s.tilt)) > 0):
                obc = s.Outside_Boundary_Condition.lower()
                if obc in ["outdoors", "ground"]:
                    iscore = False
                    break
        except BadEPFieldError:
            pass  # pass surfaces that don't have an OBC,
            # eg. InternalMass
    return iscore



################################################################################
# Module: archetypal.template
# Description:
# License: MIT, see full license in LICENSE.txt
# Web: https://github.com/samuelduchesne/archetypal
################################################################################

import collections
import logging as lg
import time
from itertools import chain, repeat

import networkx
from path import Path
from sigfig import round
from tqdm import tqdm
from validator_collection import validators

from archetypal.template.dhw import DomesticHotWaterSetting
from archetypal.template.materials.material_layer import MaterialLayer
from archetypal.template.schedule import YearSchedulePart
from archetypal.template.structure import MassRatio, StructureInformation
from archetypal.template.umi_base import UmiBase
from archetypal.template.window_setting import WindowSetting
from archetypal.template.zonedefinition import ZoneDefinition
from archetypal.utils import log, reduce


class BuildingTemplate(UmiBase):
    """Main class supporting the definition of a single building template.

    .. image:: ../images/template/buildingtemplate.png
    """

    __slots__ = (
        "_partition_ratio",
        "_lifespan",
        "_core",
        "_perimeter",
        "_structure_definition",
        "_window_setting",
        "_default_window_to_wall_ratio",
        "_year_from",
        "_year_to",
        "_country",
        "_climate_zone",
        "_authors",
        "_author_emails",
        "_version",
    )

    def __init__(
        self,
        Name,
        Core=None,
        Perimeter=None,
        Structure=None,
        Windows=None,
        Lifespan=60,
        PartitionRatio=0.35,
        DefaultWindowToWallRatio=0.4,
        YearFrom=None,
        YearTo=None,
        Country=None,
        ClimateZone=None,
        Authors=None,
        AuthorEmails=None,
        Version="v1.0",
        **kwargs,
    ):
        """Initialize a :class:`BuildingTemplate` object with the following
        attributes:

        Args:
            Core (ZoneDefinition): The Zone object defining the core zone. see
                :class:`Zone` for more details.
            Perimeter (ZoneDefinition): The Zone object defining the perimeter zone. see
                :class:`Zone` for more details.
            Structure (StructureInformation): The StructureInformation object
                defining the structural properties of the template.
            Windows (WindowSetting): The WindowSetting object defining the
                window properties of the object.
            Lifespan (float): The projected lifespan of the building template in
                years. Used in various calculations such as embodied energy.
            PartitionRatio (float): The number of lineal meters of partitions
                (Floor to ceiling) present in average in the building floor plan
                by m2.
            DefaultWindowToWallRatio (float): The default Window to Wall Ratio
                (WWR) for this template (same for all orientations). Number
                between 0 and 1.
            YearFrom (int): Start year for range.
            YearTo (int): End year for range.
            Country (list of str): alpha-3 Country Code.
            ClimateZone (list of str): ANSI/ASHRAE/IESNA Standard 90.1 International
                Climatic Zone. eg. "5A"
            Authors (list of str): Authors of this template
            AuthorEmails (list of str): Contact information.
            Version (str): Version number.
            **kwargs: other optional keywords passed to other constructors.
        """
        super(BuildingTemplate, self).__init__(Name, **kwargs)
        self.PartitionRatio = PartitionRatio
        self.Lifespan = Lifespan
        self.Core = Core
        self.Perimeter = Perimeter
        self.Structure = Structure
        self.Windows = Windows
        self.DefaultWindowToWallRatio = DefaultWindowToWallRatio
        self._year_from = YearFrom  # set privately to allow validation
        self.YearTo = YearTo
        self.Country = Country if Country else []
        self.ClimateZone = ClimateZone if ClimateZone else []
        self.Authors = Authors if Authors else []
        self.AuthorEmails = AuthorEmails if AuthorEmails else []
        self.Version = Version

    @property
    def Perimeter(self):
        """Get or set the perimeter ZoneDefinition."""
        return self._perimeter

    @Perimeter.setter
    def Perimeter(self, value):
        assert isinstance(
            value, ZoneDefinition
        ), f"Expected a ZoneDefinition, not {type(value)}"
        self._perimeter = value

    @property
    def Core(self):
        """Get or set the core ZoneDefinition."""
        return self._core

    @Core.setter
    def Core(self, value):
        assert isinstance(
            value, ZoneDefinition
        ), f"Expected a ZoneDefinition, not {type(value)}"
        self._core = value

    @property
    def Structure(self):
        """Get or set the StructureInformation."""
        return self._structure_definition

    @Structure.setter
    def Structure(self, value):
        assert isinstance(
            value, StructureInformation
        ), f"Expected a StructureInformation, not {type(value)}"
        self._structure_definition = value

    @property
    def Windows(self):
        """Get or set the WindowSetting."""
        return self._window_setting

    @Windows.setter
    def Windows(self, value):
        assert isinstance(
            value, WindowSetting
        ), f"Expected a WindowSetting, not {type(value)}"
        self._window_setting = value

    @property
    def DefaultWindowToWallRatio(self):
        """Get or set the DefaultWindowToWallRatio [-]."""
        return self._default_window_to_wall_ratio

    @DefaultWindowToWallRatio.setter
    def DefaultWindowToWallRatio(self, value):
        self._default_window_to_wall_ratio = validators.float(
            value, minimum=0, maximum=1
        )

    @property
    def Lifespan(self):
        """Get or set the building life span [years]."""
        return self._lifespan

    @Lifespan.setter
    def Lifespan(self, value):
        self._lifespan = validators.integer(value, minimum=True, coerce_value=True)

    @property
    def PartitionRatio(self):
        """Get or set the partition ratio [-]."""
        return self._partition_ratio

    @PartitionRatio.setter
    def PartitionRatio(self, value):
        self._partition_ratio = validators.float(value, minimum=0)

    @property
    def YearFrom(self):
        """Get or set the YearFrom [int]."""
        return self._year_from

    @YearFrom.setter
    def YearFrom(self, value):
        self._year_from = validators.integer(
            value, coerce_value=True, maximum=self.YearTo, allow_empty=True
        )

    @property
    def YearTo(self):
        """Get or set the YearTo [int]."""
        return self._year_to

    @YearTo.setter
    def YearTo(self, value):
        self._year_to = validators.integer(
            value, coerce_value=True, minimum=self.YearFrom, allow_empty=True
        )

    @property
    def Country(self):
        """Get or set the list of alpha-3 country codes [list]."""
        return self._country

    @Country.setter
    def Country(self, value):
        self._country = validators.iterable(value, allow_empty=True)

    @property
    def ClimateZone(self):
        """Get or set the list of climatic zones [list]."""
        return self._climate_zone

    @ClimateZone.setter
    def ClimateZone(self, value):
        self._climate_zone = validators.iterable(value, allow_empty=True)

    @property
    def Authors(self):
        """Get or set the list of authors [list]."""
        return self._authors

    @Authors.setter
    def Authors(self, value):
        self._authors = validators.iterable(value, allow_empty=True)

    @property
    def AuthorEmails(self):
        """Get or set list of author emails [list]."""
        return self._author_emails

    @AuthorEmails.setter
    def AuthorEmails(self, value):
        self._author_emails = validators.iterable(value, allow_empty=True)

    @property
    def Version(self):
        """Get or set the template version [str]."""
        return self._version

    @Version.setter
    def Version(self, value):
        self._version = validators.string(value, coerce_value=True)

    @classmethod
    def from_dict(
        cls,
        data,
        zone_definitions,
        structure_definitions,
        window_settings,
        schedules,
        window_constructions,
        **kwargs,
    ):
        """Create an BuildingTemplate from a dictionary.

        Args:
            data (dict): The python dictionary.
            zone_definitions (dict): A dictionary of ZoneDefinition objects with their
                id as keys.
            structure_definitions (dict): A dictionary of StructureInformation with
                their id as keys.
            window_settings (dict): A dictionary of WindowSetting objects with their
                id as keys.
            schedules (dict): A dictionary of UmiSchedule with their id as keys.
            window_constructions (dict): A dictionary of WindowConstruction objects
                with their id as keys.
            **kwargs: keywords passed to the constructor.

        .. code-block:: python

            {
              "Core": {
                "$ref": "178"
              },
              "Lifespan": 60,
              "PartitionRatio": 0.3,
              "Perimeter": {
                "$ref": "178"
              },
              "Structure": {
                "$ref": "64"
              },
              "Windows": {
                "$ref": "181"
              },
              "DefaultWindowToWallRatio": 0.4,
              "YearFrom": 0,
              "YearTo": 0,
              "Country": [
                "USA"
              ],
              "ClimateZone": [
                "5A"
              ],
              "Authors": [
                "Carlos Cerezo"
              ],
              "AuthorEmails": [
                "ccerezo@mit.edu"
              ],
              "Version": "v1.0",
              "Category": "Residential and Lodging",
              "Comments": "Base building definition for MIT 4433",
              "DataSource": "MIT_SDL",
              "Name": "B_Res_0_WoodFrame"
            }

        """
        core = zone_definitions[data.pop("Core")["$ref"]]
        perim = zone_definitions[data.pop("Perimeter")["$ref"]]
        structure = structure_definitions[data.pop("Structure")["$ref"]]
        window_data = data.pop("Windows")
        try:
            window = window_settings[window_data["$ref"]]
        except KeyError:
            window = WindowSetting.from_dict(
                window_data, schedules, window_constructions
            )

        return cls(
            Core=core,
            Perimeter=perim,
            Structure=structure,
            Windows=window,
            **data,
            **kwargs,
        )

    @classmethod
    def from_idf(cls, idf, **kwargs):
        """Create a BuildingTemplate from an IDF object.

        Args:
            idf (IDF):
            **kwargs:
        """
        # initialize empty BuildingTemplate
        name = kwargs.pop("Name", Path(idf.idfname).basename().splitext()[0])

        epbunch_zones = idf.idfobjects["ZONE"]
        zones = [
            ZoneDefinition.from_epbunch(ep_zone, allow_duplicates=True, **kwargs)
            for ep_zone in tqdm(epbunch_zones, desc=f"Creating UMI objects for {name}")
        ]
        # do core and Perim zone reduction
        bt = cls.reduced_model(name, zones, **kwargs)

        if not bt.Core.DomesticHotWater or not bt.Perimeter.DomesticHotWater:
            dhw = DomesticHotWaterSetting.whole_building(idf)
            if not bt.Core.DomesticHotWater:
                bt.Core.DomesticHotWater = dhw
            if not bt.Perimeter.DomesticHotWater:
                bt.Perimeter.DomesticHotWater = dhw

        bt.Comments = "\n".join(
            [
                "WWR calculated for original model: ",
                idf.wwr().to_string(),
                "where East=90, South=180, West=270, North=0\n",
            ]
        )

        bt.PartitionRatio = idf.partition_ratio

        return bt

    @classmethod
    def reduced_model(cls, name, zones, **kwargs):
        """Create reduced BuildingTemplate from list of ZoneDefinitions.

        Args:
            name (str): The name of the building template.
            zones (list of ZoneDefinition): A list of zone definition objects to
                reduce. At least one must be a perimeter zone (ZoneDefinition.is_core is
                False).
            **kwargs: keywords passed to the class constructor.

        Returns:
            BuildingTemplate: The reduced BuildingTemplate.
        """
        # reduce list of perimeter zones

        log("Initiating complexity reduction...")
        start_time = time.time()

        zone: ZoneDefinition
        cores = list(
            chain.from_iterable(
                [
                    list(repeat(zone.duplicate(), zone.multiplier))
                    for zone in zones
                    if zone.is_core
                ]
            )
        )
        perimeters = list(
            chain.from_iterable(
                [
                    list(repeat(zone.duplicate(), zone.multiplier))
                    for zone in zones
                    if not zone.is_core
                ]
            )
        )
        assert (
            len(perimeters) >= 1
        ), "Building complexity reduction must have at least one perimeter zone."

        Core = None
        # reduce list of core zones
        if cores:
            Core = reduce(
                ZoneDefinition.combine,
                tqdm(
                    cores,
                    desc=f"Reducing core zones in {name}",
                ),
            )
            Core.Name = f"{name}_ZoneDefinition_Core"  # set name

        Perimeter = None
        if perimeters:
            Perimeter = reduce(
                ZoneDefinition.combine,
                tqdm(
                    zones,
                    desc=f"Reducing perimeter zones in {name}",
                ),
            )
            Perimeter.Name = f"{name}_ZoneDefinition_Perimeter"

        # If all perimeter zones, assign self.Perimeter to core.
        if not Core:
            Core = Perimeter
            Core.Name = f"{name}_ZoneDefinition"  # rename as both core/perim

        # resolve StructureInformation and WindowSetting
        structure = StructureInformation(
            MassRatios=[MassRatio.generic()],
            Name=name + "_StructureDefinition",
        )

        # assign generic window if None
        if Perimeter.Windows is None:
            # create generic window
            Perimeter.Windows = WindowSetting.generic(Name="Generic Window")
            kwargs.setdefault("DefaultWindowToWallRatio", 0)

        log(
            f"Equivalent core zone has an area of {Core.area:,.0f} m2",
            level=lg.DEBUG,
        )
        log(
            f"Equivalent perimeter zone has an area of {Perimeter.area:,.0f} m2",
            level=lg.DEBUG,
        )
        log(
            f"Completed model complexity reduction for BuildingTemplate '{name}' "
            f"in {time.time() - start_time:,.2f}"
        )
        return cls(
            name,
            Core=Core,
            Perimeter=Perimeter,
            Windows=Perimeter.Windows,
            Structure=structure,
            **kwargs,
        )

    def _graph_reduce(self, G):
        """Using the depth first search algorithm, iterate over the zone
        adjacency graph and compute the equivalent zone yielded by the
        'addition' of two consecutive zones.

        'Adding' two zones together means both zones properties are
        weighted-averaged by zone area. All dependent objects implement the
        :func:`operator.add` method.

        Args:
            G (archetypal.zone_graph.ZoneGraph):

        Returns:
            ZoneDefinition: The reduced zone
        """
        if len(G) < 1:
            log("No zones for building graph %s" % G.name)
            return None
        else:
            log("starting reduce process for building %s" % self.Name)
            start_time = time.time()

            # start from the highest degree node
            subgraphs = sorted(
                (G.subgraph(c) for c in networkx.connected_components(G)),
                key=len,
                reverse=True,
            )
            from functools import reduce
            from operator import add

            bundle_zone = reduce(
                add,
                [zone for subG in subgraphs for name, zone in subG.nodes(data="zone")],
            )

            log(
                f"completed zone reduction for zone '{bundle_zone.Name}' "
                f"in building '{self.Name}' in {time.time() - start_time:,.2f} seconds"
            )
            return bundle_zone

    def to_dict(self):
        """Return BuildingTemplate dictionary representation."""
        self.validate()  # Validate object before trying to get json format

        data_dict = collections.OrderedDict()

        data_dict["Core"] = self.Core.to_ref()
        data_dict["Lifespan"] = self.Lifespan
        data_dict["PartitionRatio"] = round(self.PartitionRatio, 2)
        data_dict["Perimeter"] = self.Perimeter.to_ref()
        data_dict["Structure"] = self.Structure.to_ref()
        data_dict["Windows"] = self.Windows.to_ref()
        data_dict["Category"] = validators.string(self.Category, allow_empty=True)
        data_dict["Comments"] = validators.string(self.Comments, allow_empty=True)
        data_dict["DataSource"] = self.DataSource
        data_dict["Name"] = self.Name
        data_dict["YearFrom"] = self.YearFrom
        data_dict["YearTo"] = self.YearTo
        data_dict["Country"] = self.Country
        data_dict["ClimateZone"] = self.ClimateZone
        data_dict["Authors"] = self.Authors
        data_dict["AuthorEmails"] = self.AuthorEmails
        data_dict["Version"] = self.Version

        return data_dict

    def validate(self):
        """Validate object and fill in missing values."""
        return self

    def get_unique(self):
        """Replace recursively every objects with the first equivalent object."""

        def recursive_replace(umibase):
            for key, obj in umibase.mapping().items():
                if isinstance(
                    obj, (UmiBase, MaterialLayer, YearSchedulePart, MassRatio)
                ):
                    recursive_replace(obj)
                    setattr(umibase, key, obj.get_unique())
                elif isinstance(obj, list):
                    [
                        recursive_replace(obj)
                        for obj in obj
                        if isinstance(
                            obj, (UmiBase, MaterialLayer, YearSchedulePart, MassRatio)
                        )
                    ]

        recursive_replace(self)
        return self

    def mapping(self, validate=True):
        """Get a dict based on the object properties, useful for dict repr.

        Args:
            validate (bool): If True, try to validate object before returning the
                mapping.
        """
        if validate:
            self.validate()

        return dict(
            Core=self.Core,
            Lifespan=self.Lifespan,
            PartitionRatio=self.PartitionRatio,
            Perimeter=self.Perimeter,
            Structure=self.Structure,
            Windows=self.Windows,
            Category=self.Category,
            Comments=self.Comments,
            DataSource=self.DataSource,
            Name=self.Name,
            YearFrom=self.YearFrom,
            YearTo=self.YearTo,
            Country=self.Country,
            ClimateZone=self.ClimateZone,
            Authors=self.Authors,
            AuthorEmails=self.AuthorEmails,
            Version=self.Version,
        )

    def get_ref(self, ref):
        """Get item matching reference id.

        Args:
            ref:
        """
        return next(
            iter(
                [
                    value
                    for value in BuildingTemplate.CREATED_OBJECTS
                    if value.id == ref["$ref"]
                ]
            ),
            None,
        )

    def __hash__(self):
        """Return the hash value of self."""
        return hash(
            (self.__class__.__name__, getattr(self, "Name", None), self.DataSource)
        )

    def __eq__(self, other):
        """Assert self is equivalent to other."""
        if not isinstance(other, BuildingTemplate):
            return NotImplemented
        else:
            return all(
                [
                    self.Core == other.Core,
                    self.Perimeter == other.Perimeter,
                    self.Structure == other.Structure,
                    self.Windows == other.Windows,
                    self.Lifespan == other.Lifespan,
                    self.PartitionRatio == other.PartitionRatio,
                    self.DefaultWindowToWallRatio == other.DefaultWindowToWallRatio,
                    self.YearFrom == other.YearFrom,
                    self.YearTo == other.YearTo,
                    self.Country == other.Country,
                    self.ClimateZone == other.ClimateZone,
                    self.Authors == other.Authors,
                    self.AuthorEmails == other.AuthorEmails,
                    self.Version == other.Version,
                ]
            )
